@@ -2,7 +2,7 @@
  * PL/R - PostgreSQL support for R as a
  *	      procedural language (PL)
  *
- * Copyright (c) 2003, 2004 by Joseph E. Conway
+ * Copyright (c) 2003-2006 by Joseph E. Conway
  * ALL RIGHTS RESERVED
  * 
  * Joe Conway <mail@joeconway.com>
@@ -31,12 +31,12 @@
  * plr.c - Language handler and support functions
  */
 #include "plr.h"
+#include "catalog/pg_namespace.h"
 #include "storage/ipc.h"
 #include "utils/memutils.h"
-
-#if defined(PG_VERSION_80_COMPAT) || defined(PG_VERSION_81_COMPAT)
 #include "utils/typcache.h"
-#endif /* PG_VERSION_80_COMPAT || PG_VERSION_81_COMPAT */
+
+PG_MODULE_MAGIC;
 
 /*
  * Global data
@@ -48,6 +48,9 @@ char *last_R_error_msg = NULL;
 static bool	plr_pm_init_done = false;
 static bool	plr_be_init_done = false;
 
+/* namespace OID for the PL/R language handler function */
+static Oid plr_nspOid = InvalidOid;
+
 /*
  * defines
  */
@@ -115,15 +118,15 @@ static Datum plr_trigger_handler(PG_FUNCTION_ARGS);
 static Datum plr_func_handler(PG_FUNCTION_ARGS);
 static plr_function *compile_plr_function(FunctionCallInfo fcinfo);
 static plr_function *do_compile(FunctionCallInfo fcinfo,
-							    HeapTuple procTup,
-							    plr_func_hashkey *hashkey);
+								HeapTuple procTup,
+								plr_func_hashkey *hashkey);
 static SEXP plr_parse_func_body(const char *body);
 static SEXP plr_convertargs(plr_function *function, Datum *arg, bool *argnull);
 static void plr_error_callback(void *arg);
-
-#if defined(PG_VERSION_80_COMPAT) || defined(PG_VERSION_81_COMPAT)
+static Oid getNamespaceOidFromFunctionOid(Oid fnOid);
+static bool haveModulesTable(Oid nspOid);
+static char *getModulesSql(Oid nspOid);
 static char **fetchArgNames(HeapTuple procTup, int nargs);
-#endif /* PG_VERSION_80_COMPAT || PG_VERSION_81_COMPAT */
 
 /*
  * plr_call_handler -	This is the only visible function
@@ -194,7 +197,7 @@ load_r_cmd(const char *cmd)
 	SET_STRING_ELT(cmdSexp, 0, COPY_TO_USER_STRING(cmd));
 	PROTECT(cmdexpr = R_PARSEVECTOR(cmdSexp, -1, &status));
 	if (status != PARSE_OK) {
-	    UNPROTECT(2);
+		UNPROTECT(2);
 		if (last_R_error_msg)
 			ereport(ERROR,
 					(errcode(ERRCODE_DATA_EXCEPTION),
@@ -360,6 +363,7 @@ plr_load_modules(MemoryContext	plr_SPI_context)
 	int				i;
 	int				fno;
 	MemoryContext	oldcontext;
+	char		   *modulesSql;
 
 	/* switch to SPI memory context */
 	oldcontext = MemoryContextSwitchTo(plr_SPI_context);
@@ -367,19 +371,18 @@ plr_load_modules(MemoryContext	plr_SPI_context)
 	/*
 	 * Check if table plr_modules exists
 	 */
-	spi_rc = SPI_exec("select 1 from pg_catalog.pg_class "
-					  "where relname = 'plr_modules'", 1);
-	SPI_freetuptable(SPI_tuptable);
-
-	if (spi_rc != SPI_OK_SELECT)
-		/* internal error */
-		elog(ERROR, "plr_init_load_modules: select from pg_class failed");
-	if (SPI_processed == 0)
+	if (!haveModulesTable(plr_nspOid))
 		return;
 
+	/* plr_modules table exists -- get SQL code extract table's contents */
+	modulesSql = getModulesSql(plr_nspOid);
+
 	/* Read all the row's from it in the order of modseq */
-	spi_rc = SPI_exec("select modseq, modsrc from plr_modules " \
-					  "order by modseq", 0);
+	spi_rc = SPI_exec(modulesSql, 0);
+
+	/* modulesSql no longer needed -- cleanup */
+	pfree(modulesSql);
+
 	if (spi_rc != SPI_OK_SELECT)
 		/* internal error */
 		elog(ERROR, "plr_init_load_modules: select from plr_modules failed");
@@ -435,6 +438,9 @@ plr_init_all(Oid funcid)
 		/* load "builtin" R functions */
 		plr_load_builtins(funcid);
 
+		/* obtain & store namespace OID of PL/R language handler */
+		plr_nspOid = getNamespaceOidFromFunctionOid(funcid);
+
 		/* try to load procedures from plr_modules */
 		plr_load_modules(plr_SPI_context);
 
@@ -559,7 +565,7 @@ plr_trigger_handler(PG_FUNCTION_ARGS)
 
 	dims[0] = trigdata->tg_trigger->tgnargs;
 	lbs[0] = 1;
-	array = construct_md_array(dvalues, ndims, dims, lbs,
+	array = construct_md_array(dvalues, NULL, ndims, dims, lbs,
 								TEXTOID, -1, false, 'i');
 
 	arg[8] = PointerGetDatum(array);
@@ -1136,7 +1142,7 @@ do_compile(FunctionCallInfo fcinfo,
 
 		/* trigger procedure has fixed args */
 		appendStringInfo(proc_internal_args,
-						"pg.tg.name,pg.tg.relid,pg.tg.relname,pg.tg.when," \
+						"pg.tg.name,pg.tg.relid,pg.tg.relname,pg.tg.when,"
 						"pg.tg.level,pg.tg.op,pg.tg.new,pg.tg.old,pg.tg.args");
 	}
 
@@ -1203,7 +1209,7 @@ plr_parse_func_body(const char *body)
 	PROTECT(fun = VECTOR_ELT(R_PARSEVECTOR(rbody, -1, &status), 0));
 	if (status != PARSE_OK)
 	{
-	    UNPROTECT(2);
+		UNPROTECT(2);
 		if (last_R_error_msg)
 			ereport(ERROR,
 					(errcode(ERRCODE_DATA_EXCEPTION),
@@ -1336,7 +1342,6 @@ plr_error_callback(void *arg)
 		errcontext("In PL/R function %s", (char *) arg);
 }
 
-#if defined(PG_VERSION_80_COMPAT) || defined(PG_VERSION_81_COMPAT)
 /*
  * Fetch the argument names, if any, from the proargnames field of the
  * pg_proc tuple.  Results are palloc'd.
@@ -1363,7 +1368,7 @@ fetchArgNames(HeapTuple procTup, int nargs)
 
 	deconstruct_array(DatumGetArrayTypeP(argnamesDatum),
 					  TEXTOID, -1, false, 'i',
-					  &elems, &nelems);
+					  &elems, NULL, &nelems);
 
 	if (nelems != nargs)		/* should not happen */
 		elog(ERROR, "proargnames must have the same number of elements as the function has arguments");
@@ -1375,4 +1380,101 @@ fetchArgNames(HeapTuple procTup, int nargs)
 
 	return result;
 }
-#endif /* PG_VERSION_80_COMPAT || PG_VERSION_81_COMPAT */
+
+/*
+ * getNamespaceOidFromFunctionOid - Returns the OID of the namespace for the
+ * language handler function for the postgresql function with the OID equal
+ * to the input argument.
+ */
+static Oid
+getNamespaceOidFromFunctionOid(Oid fnOid)
+{
+	HeapTuple			procTuple;
+	HeapTuple			langTuple;
+	Form_pg_proc		procStruct;
+	Form_pg_language	langStruct;
+	Oid					langOid;
+	Oid					hfnOid;
+	Oid					nspOid;
+
+	/* Lookup the pg_proc tuple for the called function by OID */
+	procTuple = SearchSysCache(PROCOID, ObjectIdGetDatum(fnOid), 0, 0, 0);
+
+	if (!HeapTupleIsValid(procTuple))
+		/* internal error */
+		elog(ERROR, "cache lookup failed for function %u", fnOid);
+
+	procStruct = (Form_pg_proc) GETSTRUCT(procTuple);
+	langOid = procStruct->prolang;
+	ReleaseSysCache(procTuple);
+
+	/* Lookup the pg_language tuple by OID */
+	langTuple = SearchSysCache(LANGOID, ObjectIdGetDatum(langOid), 0, 0, 0);
+	if (!HeapTupleIsValid(langTuple))
+		/* internal error */
+		elog(ERROR, "cache lookup failed for language %u", langOid);
+
+	langStruct = (Form_pg_language) GETSTRUCT(langTuple);
+	hfnOid = langStruct->lanplcallfoid;
+	ReleaseSysCache(langTuple);
+
+	/* Lookup the pg_proc tuple for the language handler by OID */
+	procTuple = SearchSysCache(PROCOID, ObjectIdGetDatum(hfnOid), 0, 0, 0);
+
+	if (!HeapTupleIsValid(procTuple))
+		/* internal error */
+		elog(ERROR, "cache lookup failed for function %u", hfnOid);
+
+	procStruct = (Form_pg_proc) GETSTRUCT(procTuple);
+	nspOid = procStruct->pronamespace;
+	ReleaseSysCache(procTuple);
+
+	return nspOid;
+}
+
+/*
+ * haveModulesTable(Oid) -- Check if table plr_modules exists in the namespace
+ * designated by the OID input argument.
+ */
+static bool
+haveModulesTable(Oid nspOid)
+{
+	StringInfo		sql = makeStringInfo();
+	char		   *sql_format = "SELECT NULL "
+								 "FROM pg_catalog.pg_class "
+								 "WHERE "
+								 "relname = 'plr_modules' AND "
+								 "relnamespace = %u";
+    int  spiRc;
+
+	appendStringInfo(sql, sql_format, nspOid);
+
+	spiRc = SPI_exec(sql->data, 1);
+	if (spiRc != SPI_OK_SELECT)
+		/* internal error */
+		elog(ERROR, "haveModulesTable: select from pg_class failed");
+
+	return SPI_processed == 1;
+}
+
+/*
+ * getModulesSql(Oid) - Builds and returns SQL needed to extract contents from
+ * plr_modules table.  The table must exist in the namespace designated by the
+ * OID input argument.  Results are ordered by the "modseq" field.
+ *
+ * IMPORTANT: return value must be pfree'd
+ */
+static char *
+getModulesSql(Oid nspOid)
+{
+	StringInfo		sql = makeStringInfo();
+	char		   *sql_format = "SELECT modseq, modsrc "
+								 "FROM %s "
+								 "ORDER BY modseq";
+
+	appendStringInfo(sql, sql_format,
+					 quote_qualified_identifier(get_namespace_name(nspOid),
+												"plr_modules"));
+
+    return sql->data;
+}
