@@ -191,7 +191,7 @@ create or replace function pg_quote_ident(text) returns text as 'pg.quoteident(a
 select pg_quote_ident('Hello World');
 
 create or replace function pg_quote_literal(text) returns text as 'pg.quoteliteral(arg1)' language 'plr';
-select pg_quote_literal('Hello\'World');
+select pg_quote_literal('Hello''World');
 
 create or replace function test_spi_t(text) returns text as '(pg.spi.exec(arg1))[[1]]' language 'plr';
 select test_spi_t('select oid, typname from pg_type where typname = ''oid'' or typname = ''text''');
